#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>
#include <assert.h>

int mediane (int a, int b, int c) {
  if ((a <= b && b <= c) || (c <= b && b <= a))
    return b;
  if ((a <= c && c <= b) || (b <= c && c <= a))
    return c;
  return a;
}

bool bissextile (int annee) {
  return (annee % 4 == 0) && !((annee % 100 == 0) && (annee % 400 != 0));
}

void compte_for (int n) { // on suppose n >= 0
  // fonctionne galement si n == 0
  for (int i = -n; i < n; i++) {
    printf("%d ", i);
  }
  printf("%d\n", n);
}

void compte_while (int n) { // on suppose n >= 0
  int i = -n;
  while (i < n) {
    printf("%d ", i);
    i++;
  }
  printf("%d\n", n);
}

bool cherche (int* t, int n, int x) {
  for (int i = 0; i < n; i++) {
    if (t[i] == x) return true;
  }
  return false;
}

int* tableau_aleatoire (int n) {
  int* t = malloc(n * sizeof(int));
  for (int i = 0; i < n; i++) {
    t[i] = rand();
  }
  return t;
}

// tri par slection en place rcursif
void tri_selection (int* t, int n) {
  if (n != 0) {
    // recherche du max
    int i_max = 0;
    for (int i = 1; i < n; i++) {
      if (t[i] > t[i_max]) i_max = i;
    }
    int tmp = t[i_max];
    t[i_max] = t[n - 1];
    t[n - 1] = tmp;
    tri_selection(t, n - 1);
  }
}

bool cherche_dicho (int* t, int n, int x) {
  int i = 0, j = n;
  while (i < j) {
    int m = (i + j) / 2;
    if (t[m] == x)
      return true;
    else if (t[m] > x)
      j = m;
    else
      i = m + 1;
  }
  return t[i] == x;
}

int maximum (int* t, int n) {
  int i_max = 0;
  for (int i = 1; i < n; i++) {
    if (t[i] > t[i_max]) i_max = i;
  }
  return t[i_max];
}

int max_local (int* t, int n) {
  for (int i = 0; i < n; i++) {
    if (i == 0 && t[i] >= t[i + 1]) return i;
    if (i == n - 1 && t[i] >= t[i - 1]) return i;
    if (t[i] >= t[i - 1] && t[i] >= t[i + 1]) return i;
  }
  // pour viter les warnings
  return -1;
}

int max_local_dicho (int* t, int n) {
  int i = 0, j = n;
  while (i < j) {
    int m = (i + j) / 2;
    if (t[m] < t[m + 1]) i = m + 1;
    else if (t[m] < t[m - 1]) j = m;
    else {
      i = m;
      break;
    }
  }
  return i;
}

// on procde par liste doublement chane
struct maillon {
  int val;
  struct maillon* prev;
  struct maillon* next;
};
// on garde des pointeurs sur le dbut et la fin
typedef struct {
  struct maillon* begin;
  struct maillon* end;
} queue;

queue create () {
  queue q = {NULL, NULL};
  return q;
}

void add (int x, queue* q) {
  struct maillon* new_end = malloc(sizeof(struct maillon));
  new_end->val = x;
  new_end->prev = q->end;
  new_end->next = NULL;
  // on met  jour le dernier maillon de la file s'il existe
  if (q->end != NULL) q->end->next = new_end;
  q->end = new_end;
  // si la file tait vide
  if (q->begin == NULL) q->begin = new_end;
}

int take (queue* q) {
  assert(q->begin != NULL);
  int x = q->begin->val;
  struct maillon* old_begin = q->begin;
  // si la file ne contient qu'un seul lment
  if (q->begin->next == NULL) {
    q->begin = NULL;
    q->end = NULL;
  } else {
    q->begin = q->begin->next;
    q->begin->prev = NULL;
  }
  free(old_begin);
  return x;
}

int peek (queue* q) {
  assert(q->begin != NULL);
  return q->begin->val;
}

bool is_empty (queue* q) {
  return q->begin == NULL;
}

int main () {
  // test de la fonction mediane
  printf("Mdiane de 4, 6 et 8: %d.\n", mediane(8, 4, 6));

  // test de la fonction bissextile
  printf("Entrez une anne: ");
  int an;
  scanf("%d", &an);
  if (bissextile(an))
    printf("L'anne %d est bissextile.\n", an);
  else
    printf("L'anne %d n'est pas bissextile.\n", an);

  // test des fonctions compte_for et compte_while
  compte_for(5);
  compte_while(3);

  // test des fonctions cherche et tableau_aleatoire
  srand(time(NULL));
  int n = 10;
  int* t = tableau_aleatoire(n);
  for (int i = 0; i < n; i++) {
    printf("%d ", t[i]);
  }
  printf("\n");
  int x = rand();
  if (cherche(t, n, x))
    printf("%d est dans le tableau\n", x);
  else
    printf("%d n'est pas dans le tableau\n", x);

  // tri et recherche dichotomique
  tri_selection(t, n);
  if (cherche_dicho(t, n, x))
    printf("%d est dans le tableau\n", x);
  else
    printf("%d n'est pas dans le tableau\n", x);

  // maximum global
  printf("%d est un maximum global\n", maximum(t, n));

  // maximum local
  for (int i = 0; i < n; i++) {
    t[i] = rand();
    printf("%d ", t[i]);
  }
  printf("\n");
  printf("%d est un maximum local\n", t[max_local(t, n)]);
  printf("%d est un maximum local\n", t[max_local_dicho(t, n)]);

  // on n'oublie pas de librer la mmoire
  free(t);

  // test des files
  queue q = create();
  for (int i = 0; i < n; i++) {
    add(i, &q);
  }
  printf("Au dbut: %d\n", peek(&q));
  while (!is_empty(&q)) {
    printf("%d ", take(&q));
  }
  printf("\n");
  /* pour retourner une file, on change tous les pointeurs prev et next, et on
     change les pointeurs begin et end */

  return 0;
}
