let val_abs_int (x : int) : int =
  match x with
  | y when y <= 0 -> - y
  | _ -> x

(* ajouter des points pour les flottants *)
let val_abs_float (x : float) : float =
  match x with
  | y when y <= 0. -> -. y
  | _ -> x

let quadruple (x : int) : int =
  let double (x : int) : int = 2 * x in
  double (double x)

(* int -> int *)
let g1 (x : int) : int = 0
(* int * int -> int *)
let g2 ((x, y) : int * int) : int = x + y
(* int -> int -> int *)
let g3 (x : int) (y : int) : int = x * y
(* int -> int * int *)
let g4 (x : int) : int * int = (x, x + 1)
(* (int -> int) -> int *)
let g5 (f : int -> int) : int = 2 * (f 0)
(* Remarque: les annotations de types ici sont inutiles car j'ai fait attention
   à bien contraindre le type des arguments et le type de retour en utilisant
   des constantes et fonctions entières *)

let f1 (f : float -> float) : float = (f 5. +. f (-. 5.)) /. 2.

let f2 ((f : float -> float), (x : float)) : float = (f x) ** 2.

(* (float -> float) -> float -> float *)
let f3 (f : float -> float) : float -> float = function x -> f2 (f, x)

(* (float -> float) -> float -> float *)
let f4 (f : float -> float) : float -> float =
  function x -> (f (2. *. x) : float)

let somme (f : int -> int) (g : int -> int) : int -> int =
  function (x : int) -> f x + g x

let produit (f : int -> int) (g : int -> int) : int -> int =
  function (x : int) -> f x * g x

let produit_ext (f : int -> int) (a : int) (x : int) : int = a * f x

let derive (f : float -> float) (eps : float) (x : float) : float =
  (f (x +. eps) -. f x) /. eps

(* à peu près 14 *)
let _ =
  let f (x : float) : float = x *. x *. x +. 2. *. x +. 1. in
  derive f (10. ** (-. 5.)) 2.

let rec valide (l : 'a list) : bool =
  match l with
  | [] -> true
  | t :: q when List.mem t q -> false
  | _ :: q -> valide q

(* on code List.mem ainsi *)
let rec mem (x : 'a) (l : 'a list) : bool =
  match l with
  | [] -> false
  | t :: q -> t = x || mem x q

let rec unique (l : 'a list) : 'a list =
  match l with
  | [] -> []
  | t :: q when List.mem t q -> unique q
  | t :: q -> t :: unique q

let union (l1 : 'a list) (l2 : 'a list) : 'a list = unique (l1 @ l2)

(* on code la concaténation @ ainsi *)

let rec concat (l1 : 'a list) (l2 : 'a list) : 'a list =
  match l1 with
  | [] -> l2
  | t :: q -> t :: (concat q l2)

let rec intersection (l1 : 'a list) (l2 : 'a list) : 'a list =
  match l1 with
  | [] -> []
  | t :: q when List.mem t l2 -> t :: intersection q l2
  | _ :: q -> intersection q l2

(* en utilisant la bibliothèque standard: *)
let intersection (l1 : 'a list) (l2 : 'a list) : 'a list =
  List.filter (function t -> List.mem t l2) l1

(* List.filter ne garde que les éléments qui vérifient le prédicat
   on la code ainsi *)
let rec filter (f : 'a -> bool) (l : 'a list) : 'a list =
  match l with
  | [] -> []
  | t :: q when f t -> t :: filter f q
  | _ :: q -> filter f q

let rec inclus (l1 : 'a list) (l2 : 'a list) : bool =
  match l1 with
  | [] -> true
  | t :: q when List.mem t l2 -> inclus q l2
  | _ -> false

(* en utilisant la bibliothèque standard: *)
let inclus (l1 : 'a list) (l2 : 'a list) : bool =
  List.for_all (function t -> List.mem t l2) l1

(* List.for_all vérifie si tous les éléments vérifient le prédicat
   on la code ainsi *)
let rec for_all (f : 'a -> bool) (l : 'a list) : bool =
  match l with
  | [] -> true
  | t :: q -> f t && for_all f q

(* par double inclusion *)
let egal (l1 : 'a list) (l2 : 'a list) : bool =
  inclus l1 l2 && inclus l2 l1

type 'a bin_tree = Empty | Node of 'a * 'a bin_tree * 'a bin_tree

let rec height (t : 'a bin_tree) : int =
  match t with
  | Empty -> -1
  | Node (_, l, r) -> 1 + max (height l) (height r)

let infix_dfs (t : 'a bin_tree) : 'a list =
  (* on utilise une fonction auxiliaire pour améliorer la complexité
     aux t acc renvoie le parcours infixe de t concaténé avec acc
     l'usage de l'accumulateur évite les concaténations de listes (linéaires en
     la taille de la liste de gauche) et les remplace par l'usage du
     constructeur :: *)
  let rec aux (t : 'a bin_tree) (acc : 'a list) : 'a list =
    match t with
    | Empty -> acc
    | Node (x, l, r) -> aux l (x :: aux r acc)
  in aux t []

let is_bst (t : 'a bin_tree) : bool =
  (* il suffit de vérifier si le parcours infixe est trié *)
  let rec is_sorted (l : 'a list) : bool =
    match l with
    | [] -> true
    | [_] -> true
    | t1 :: t2 :: q -> t1 <= t2 && is_sorted (t2 :: q)
  in is_sorted (infix_dfs t)

let rec mem (x : 'a) (t : 'a bin_tree) : bool =
  match t with
  | Empty -> false
  (* on cherche du bon côté selon la comparaison de x et de la racine *)
  | Node (y, _, _) when y = x -> true
  | Node (y, l, _) when x < y -> mem x l
  | Node (_, _, r) -> mem x r

(* le max est le nœud le plus à droite *)
let rec max (t : 'a bin_tree) : 'a =
  match t with
  | Empty -> failwith "Empty tree"
  | Node (y, _, Empty) -> y
  | Node (_, _, r) -> max r

(* on descend jusqu'à la création d'une nouvelle feuille *)
let rec add (x : 'a) (t : 'a bin_tree) : 'a bin_tree =
  match t with
  | Empty -> Node (x, Empty, Empty)
  | Node (y, l, r) when x <= y -> Node (y, add x l, r)
  | Node (y, l, r) -> Node (y, l, add x r)

(* si on supprime un nœud, on veut fusionner ses deux sous-arbres *)
let rec merge (l : 'a bin_tree) (r : 'a bin_tree) : 'a bin_tree =
  match l with
  | Empty -> r
  | Node (y, ll, lr) -> Node (y, ll, merge lr r)

let rec del (x : 'a) (t : 'a bin_tree) : 'a bin_tree =
  match t with
  | Empty -> t
  | Node (y, l, r) when x < y -> Node (y, del x l, r)
  | Node (y, l, r) when x > y -> Node (y, l, del x r)
  | Node (_, l, r) -> merge l r
